/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2006 , wojtek@VoIPMan.ORG, rights assigned to Digium
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief ZIM-SMS application - added support for ZIM-SMS gateway
 * \ingroup applications
 *
 */
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <ctype.h>
#include <curl/curl.h>

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 7634 $")


#include "asterisk/lock.h"
#include "asterisk/logger.h"
#include "asterisk/options.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/config.h"
#include "asterisk/callerid.h"
#include "asterisk/utils.h"
#include "asterisk/module.h"
#include "asterisk/app.h"
#include "asterisk/zim-sms.h"



static char *app = "ZIM-SMS";

static char *tdesc = "SMS gateway via ZIM";

static char *synopsis = "Interface to ZIM SMS Services";

static char *descrip = "  ZIM-SMS(sms number|body text|subject):  ZIM-SMS allows text messages to be sent to\n"
"GSM network via ZIM.BIZ.\n"
"Two arguments are required, the destination number and the text of the message.\n"
"Subject of the message could be presentes as third, optional argument.\n"
"If not presented, the default subject from the config file will be used.\n"
"The destination number must include the country code and no + sign: ie: 16132618299\n"
"The text message will be truncated if it is more than 160 characters.\n";

STANDARD_LOCAL_USER;

CURL *curl;

#define ZIM_SMS_CONFIG "zim-sms.conf"

/* ZIM Request Struct */
typedef struct ZIM_REQUEST {
   size_t number;
   size_t date;
   size_t time;
   char* code;
   union {
  	 char* status;
  	 char *error;
   };
   int num_email_received, num_phone_received;
} ZIM_REQUEST;

ZIM_REQUEST *zim_request;

/* prototypes */
static int zim_curlwrite( void *ptr, size_t size, size_t nmemb, void *stream);
static void * zim_parse_tag(void *stag,void * etag);
static void zim_request_free(void);

/**
 * zim_request_free -  free all the memory
 */
static void zim_request_free(void) {
   free(zim_request->code);
   free(zim_request->status);
   free(zim_request);
   zim_request = NULL;
}


/* Command processing */
int ast_zim_sms(struct ast_channel *chan, void *data) {
  struct ast_config *cfg;
  char *code;
  char *key;
  char mobile[25] = "";
  char text[120]= "";
  char *Ptr1 = "";
  char *params;
  char subject[25] = "";
  char *subject_d;
  char *argv[3], *tmp;
  int argc;
  long responseCode;

  ast_log(LOG_DEBUG,"ZIM-SMS exec started\n");

  if (!data || ast_strlen_zero((char *)data)) {
    ast_log(LOG_ERROR, "ZIM-SMS requires arguments (mobile,message text)\n");
    return -1;
  }
      
  /* Load config //params */
  ast_log(LOG_DEBUG,"ZIM-SMS attempts to load the config file\n");
  cfg = ast_config_load(ZIM_SMS_CONFIG);
  if (!cfg) {
    ast_log(LOG_ERROR,"ZIM-SMS configuration file (%s) is missing",ZIM_SMS_CONFIG);
    return -1;
  }
  code = ast_variable_retrieve(cfg,"general","code");
  key = ast_variable_retrieve(cfg,"general","key");
  subject_d = ast_variable_retrieve(cfg,"general","subject");
  if(!subject_d)
	subject_d = "";
    
  if (!code) {
    ast_log(LOG_ERROR,"Code parameter is missing from %s",ZIM_SMS_CONFIG);
    return -1;
  }
  if (!key) {
    ast_log(LOG_ERROR,"Key parameter is missing from %s",ZIM_SMS_CONFIG);
    return -1;
  }
 
  
  ast_log(LOG_DEBUG,"ZIM-SMS will now parse for parameters\n");
  /* Get params */
   if (!ast_strlen_zero((char *)data)) {
 	tmp = ast_strdupa(data);
 	argc = ast_app_separate_args(tmp, '|', argv, sizeof(argv) / sizeof(argv[0]));

 	if (argc >= 2 ) {
 		if (!ast_strlen_zero(argv[0])) {
 			ast_copy_string(mobile, argv[0], sizeof(mobile));
			for (Ptr1=mobile; Ptr1 < (mobile+strlen(mobile)); Ptr1++){
                		if ( isalpha(*Ptr1) != 0){
                        		ast_log(LOG_ERROR,"ZIM-SMS mobile can only contain digits\n");
                        		ast_log(LOG_DEBUG,"ZIM-SMS trying to send message to the wrong number %s\n", mobile);
                        		return -1;
                		}
        		}
                        ast_log(LOG_DEBUG,"ZIM-SMS got the SMS number: %s\n", mobile);
		} else {
			ast_log(LOG_DEBUG,"ZIM-SMS no SMS number detected!!!");
			return -1;
		}			
 		if (!ast_strlen_zero(argv[1])) {
			int esc = 0;
 			ast_copy_string(text, argv[1], sizeof(text));
			for (Ptr1=text; Ptr1 < (text+strlen(text)); Ptr1++){
                        	if (*Ptr1 == '"' || *Ptr1 == '<' || *Ptr1 == '>')
                                	*Ptr1 = '`';
				else if (*Ptr1 == '/') { 
					esc = 1;
                                	*Ptr1 = ' ';
				} else if (*Ptr1 == 'n' && esc == 1) {
                                	*Ptr1 = ' ';
					esc = 0;
				}
        		}
		} else 
                        return -1;
		
		ast_log(LOG_DEBUG,"ZIM-SMS got the body: %s\n", text);
 		if (!ast_strlen_zero(argv[2])) {
 			ast_copy_string(subject, argv[2], sizeof(subject));
			ast_log(LOG_DEBUG,"ZIM-SMS got the subject: %s\n", subject);
		} else {
			if (subject_d) 
				ast_log(LOG_DEBUG,"ZIM-SMS using the default subject: %s\n", subject_d);
				ast_copy_string(subject, subject_d, sizeof(subject));
		}
		ast_log(LOG_DEBUG,"ZIM-SMS got the subject: %s\n", subject);
	} else 
		ast_log(LOG_ERROR,"Not enough arguments provided.");
    } else
	ast_log(LOG_ERROR,"No arguments AT ALL.");	

  ast_log(LOG_DEBUG,"ZIM-SMS (%s,%s,%s) has been prepared for delivery\n",mobile,text,subject); 

  /* Send SMS */
      
   asprintf(&params,"<?xml version='1.0' encoding='UTF-8'?>\r\n"
	"<service xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' xsi:noNamespaceSchemaLocation=' zxmlapi.xsd'>\r\n"
	"<code>%s</code>\r\n"
	"<key>%s</key>\r\n"
	"<filename>zim-sms.c</filename>\r\n"
	"<type>M</type>\r\n"
	"<message>\r\n"
		"<subject>%s</subject>\r\n"
		"<body>%s</body>\r\n"
		"<to>\r\n"
			"<atype>M</atype>\r\n"
			"<address>%s</address>\r\n"
		"</to>\r\n"
	"</message>\r\n"
	"</service>\r\n",code,key,subject,text,mobile);
  curl_easy_setopt(curl,CURLOPT_URL,"http://post2.zimsms.com/ztranspo/zxmlapi.asp");
  curl_easy_setopt(curl,CURLOPT_POSTFIELDS,params);
  curl_easy_setopt(curl,CURLOPT_SSL_VERIFYPEER,0);
  curl_easy_setopt(curl,CURLOPT_WRITEFUNCTION,zim_curlwrite);

  curl_easy_perform(curl);
  responseCode = 0;
  curl_easy_getinfo(curl,CURLINFO_HTTP_CODE,&responseCode);

  ast_log(LOG_DEBUG,"ZIM-SMS response code was %ld\n",responseCode);
  free(params);
  if (responseCode != 200) {

	ast_log(LOG_ERROR,"ZIM-SMS exec failed (%ld)\n",responseCode);
   
    zim_request_free();
    return -1;
  } else {

    if (zim_request->num_email_received == -1 && zim_request->num_phone_received == -1 ) {
        ast_log(LOG_ERROR,"ZIM-SMS Error: %s\n", zim_request->error);
        ast_log(LOG_DEBUG,"ZIM-SMS Error: %s\n", zim_request->error);
        zim_request_free();
	return -1;
	
    } else {
    	ast_log(LOG_NOTICE,"Message has been sent through ZIM-SMS.\n");
    	ast_log(LOG_DEBUG,"Message has been submitted to ZIM-SMS.\n");
    }
    zim_request_free();
    return 0;
  }
}




/**
 * zim_curlwrite - This function gets called by libcurl as soon as there is data received that needs to be saved. 
 */
static int zim_curlwrite( void *ptr, size_t size, size_t nmemb, void *stream) {
    int code_size=0;
 
    
    zim_request = (ZIM_REQUEST*)malloc(sizeof(ZIM_REQUEST));    
    zim_request->code = (char*)malloc(size*nmemb);
    zim_request->status = (char*)malloc(size*nmemb);


    /* quick check to see if this is an error */
    if (strstr((char*)ptr,"<error>") != NULL) {
	ast_log(LOG_DEBUG,"ERROR CURLWRITE");
	zim_request->error = (char*)zim_parse_tag(strstr((char*)ptr, "<error>"),strstr((char*)ptr,"</error>"));
	return 0;	
    }

    /* get number,date,time */
    sscanf((char*)strstr((char*)ptr,"<number>"),"<number>%d</number>\n",&zim_request->number);
    sscanf((char*)strstr((char*)ptr,"<date>"),"<date>%d</date>\n",&zim_request->date);
    sscanf((char*)strstr((char*)ptr,"<time>"),"<time>%d</time>\n",&zim_request->time);
    
    /* get code and status */
    zim_request->code = (char*)zim_parse_tag(strstr((char*)ptr, "<code>"),strstr((char*)ptr,"</code>"));
    zim_request->status = (char*)zim_parse_tag(strstr((char*)ptr, "<status>"),strstr((char*)ptr,"</status>"));

     
    ast_log(LOG_DEBUG,"ZIM_REQUEST->number: %d\n", zim_request->number);
    ast_log(LOG_DEBUG,"ZIM_REQUEST->date: %d\n", zim_request->date);
    ast_log(LOG_DEBUG,"ZIM_REQUEST->time: %d\n", zim_request->time);
    ast_log(LOG_DEBUG,"ZIM_REQUEST->code_size: %d\n", code_size);
    ast_log(LOG_DEBUG,"ZIM_REQUEST->code: %s\n", zim_request->code);
    ast_log(LOG_DEBUG,"ZIM_REQUEST->status: %s\n", zim_request->status);
	
    /* Ckechk for error message */
    /* Meesage Structer: X email addresses and Y phone numbers received. */
   //int num_email_send, num_phone_received;
    if (sscanf(zim_request->status, "%d email addresses and %d phone numbers received.", &zim_request->num_email_received, &zim_request->num_phone_received) == 0 ) {
	zim_request->num_email_received = -1;
	zim_request->num_phone_received = -1;
    }
    
    ast_log(LOG_DEBUG,"ZIM_REQUEST->num_email_received: %d\n", zim_request->num_email_received);
    ast_log(LOG_DEBUG,"ZIM_REQUEST->num_phone_received: %d\n", zim_request->num_phone_received);
 
    return 0;
}

/**
 * zim_parse_tag - return the cdata information for xml or some sort of tag note each tag is it on line
 * @author Pawel Pastuszak - pawelpastuszak@gmail.com
 * @param stag - address of the location of the starting tag
 * @param etag - address of the location of the ending tag
 * @return char * two the cdata element
 */
static void * zim_parse_tag(void *stag,void * etag)  {	
   int size=0,i=0;
   void *beging;
   void *val=NULL;

   beging = (char*)strstr((char*)stag,">");

   // could not find it.
   if (beging == NULL) 
	return NULL;


   // Find the size
   size = etag - beging;

   if (size <=0 ) 
	return NULL;

   val = (char*)calloc(size+1,sizeof(char*));
   beging++;
   for (i=0;i < (size-1); i++) {
      *(char*)(val + i) = *(char*)(beging +i);
   }

   return val;

}

/* Asterisk Module Interface */

//! Initialize the module
/*!
 * This function is called at module load time.  Put all code in here
 * that needs to set up your module's hardware, software, registrations,
 * etc.
 */
int load_module(void) {
  curl = curl_easy_init();
  return ast_register_application (app, ast_zim_sms, synopsis, descrip);
} 

//! Cleanup all module structures, sockets, etc
/*!
 * This is called at exit.  Any registrations and memory allocations need
 * to be unregistered and free'd here.  Nothing else will do these for you (until exit).
 * Return 0 on success, or other than 0 if there is a problem.
 */
int unload_module(void) {
  curl_easy_cleanup(curl);
  zim_request_free();
  return ast_unregister_application (app);
}

//! Provides a usecount
/*!
 * This function will be called by various parts of asterisk.  Basically, all it has
 * to do is to return a usecount when called.  You will need to maintain your usecount
 * within the module somewhere.
 */
int usecount(void) {
  return 0;
}

//! Description
/*!
 * Returns a short description of your module.
 */
char *description(void) {
  return tdesc;
}

//! Returns the ASTERISK_GPL_KEY
/*!
 * This returns the ASTERISK_GPL_KEY, signifiying that you agree to the terms of
 * the GPL stated in the ASTERISK_GPL_KEY.  Your module will not load if it does
 * not return the EXACT message, i.e.  char *key(void){return ASTERISK_GPL_KEY;}
 */
char *key(void) {
  return ASTERISK_GPL_KEY;
}

//! Reload stuff
/*!
 * This function is where any reload routines take place.  Re-read config files,
 * change signalling, whatever is appropriate on a reload.
 * Return 0 on success, and other than 0 on problem.
 */
int reload(void) {
  return 0;
}

